# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from pykeg.util.email import build_message
from pykeg.core import models as core_models
from pykeg.notification.backends.base import BaseNotificationBackend

import logging

logger = logging.getLogger('email-notification')


class EmailNotificationBackend(BaseNotificationBackend):
    def notify(self, event, user):
        logger.info('Event %s -> user %s' % (event, user))

        to_address = user.email
        if not to_address:
            logger.warning('No e-mail address available for user %s' % user)
            return

        context = {}
        template_name = None

        kbsite = core_models.KegbotSite.get()
        context['site_name'] = kbsite.title
        context['site_url'] = kbsite.base_url()
        context['settings_url'] = context['site_url'] + '/account'

        context['drink'] = event.drink
        context['keg'] = event.keg
        context['session'] = event.session

        if event.kind == core_models.SystemEvent.KEG_TAPPED:
            template_name = 'notification/email_keg_tapped.html'
            context['url'] = event.keg.full_url()
        elif event.kind == core_models.SystemEvent.KEG_ENDED:
            template_name = 'notification/email_keg_ended.html'
            context['url'] = event.keg.full_url()
        elif event.kind == core_models.SystemEvent.SESSION_STARTED:
            template_name = 'notification/email_session_started.html'
            context['url'] = event.session.short_url()
        elif event.kind == core_models.SystemEvent.KEG_VOLUME_LOW:
            template_name = 'notification/email_keg_volume_low.html'
            context['url'] = event.keg.full_url()
        else:
            logger.info('Skipping unknown event type: %s' % event.kind)
            return

        message = build_message(to_address, template_name, context)
        if not message:
            logger.warning('Aborting due to error building message.')
            return
        self.send_message(message)

    def send_message(self, message):
        logger.info('Sending message: %s' % message)
        message.send(fail_silently=True)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous utility functions."""

# Note: imports should be limited to python stdlib, since methods here
# may be used in models.py, settings.py, etc.

import logging
import pkgutil
import os
import pkg_resources
import requests
import sys
import tempfile
from threading import current_thread
from contextlib import closing
from distutils.version import StrictVersion

from redis.exceptions import RedisError

from django.core.exceptions import ImproperlyConfigured

logger = logging.getLogger(__name__)

_REQUESTS = {}


def get_version():
    try:
        return pkg_resources.get_distribution('kegbot').version
    except pkg_resources.DistributionNotFound:
        return '0.0.0'


def get_version_object():
    return StrictVersion(get_version())


def must_upgrade(installed_version, new_version):
    # Compare major and minor (only).
    return installed_version.version[:2] < new_version.version[:2]


def should_upgrade(installed_verison, new_version):
    return installed_verison < new_version


def get_user_agent():
    return 'KegbotServer/%s' % get_version()


def get_plugin_template_dirs(plugin_list):
    from django.utils import six
    if not six.PY3:
        fs_encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()

    ret = []
    for plugin in plugin_list:
        plugin_module = '.'.join(plugin.split('.')[:-1])
        pkg = pkgutil.get_loader(plugin_module)
        if not pkg:
            raise ImproperlyConfigured('Cannot find plugin "%s"' % plugin)
        template_dir = os.path.join(os.path.dirname(pkg.filename), 'templates')
        if os.path.isdir(template_dir):
            if not six.PY3:
                template_dir = template_dir.decode(fs_encoding)
            ret.append(template_dir)
    return ret


def get_current_request():
    """Retrieve the current request.

    Adapted from: http://nedbatchelder.com/blog/201008/global_django_requests.html
    """
    return _REQUESTS.get(current_thread())


def set_current_request(request):
    thr = current_thread()
    if request:
        _REQUESTS[thr] = request
    else:
        if thr in _REQUESTS:
            del _REQUESTS[thr]


def download_to_tempfile(url):
    try:
        r = requests.get(url, stream=True)
        ext = os.path.splitext(url)[1]
        fd, pathname = tempfile.mkstemp(suffix=ext)
        logger.info('Downloading file %s to path %s' % (url, pathname))
        with closing(os.fdopen(fd, 'wb')):
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    os.write(fd, chunk)
        return str(pathname)
    except requests.exceptions.RequestException as e:
        raise IOError('Could not download file: {}'.format(e))


class SuppressTaskErrors(object):
    """Suppresses certain errors that occur while scheduling tasks."""

    def __init__(self, logger=None):
        self.logger = logger if logger else logging.getLogger(__name__)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        exc_info = (exc_type, exc_val, exc_tb)
        if isinstance(exc_val, RedisError):
            self.logger.error('Error scheduling task: {}'.format(exc_val),
                exc_info=exc_info)
            return True
        return False



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Celery tasks for Foursquare."""

import os
import foursquare
import datetime
import PIL

from cStringIO import StringIO
from pykeg.celery import app
from pykeg.plugin import util
from pykeg.core.util import download_to_tempfile

MAX_CHECKIN_AGE = datetime.timedelta(minutes=30)

logger = util.get_logger(__name__)


@app.task(name='foursquare_checkin', expires=60)
def foursquare_checkin(token, venue_id, image_url=None):
    client = foursquare.Foursquare(access_token=token)
    logger.info('Checkin request: token=%s venue=%s' % (token, venue_id))

    checkin = None
    last_checkin = _get_last_checkin(client)
    if last_checkin:
        venue = last_checkin.get('venue', {}).get('id', '')
        if venue == venue_id:
            when = datetime.datetime.fromtimestamp(int(last_checkin.get('createdAt', 0)))
            if (when + MAX_CHECKIN_AGE) > datetime.datetime.now():
                logger.info('Reusing last checkin.')
                checkin = last_checkin

    if not checkin:
        logger.info('Performing fresh checkin.')
        checkin = client.checkins.add(params={'venueId': venue_id}).get('checkin')

    checkin_id = checkin.get('id')

    image_path = None
    if image_url:
        logger.info('Has image, downloading: {}'.format(image_url))
        try:
            image_path = download_to_tempfile(image_url)
        except IOError as e:
            logger.warning("Couldn't download file: {}".format(e))

    image_bytes = None
    if image_path:
        try:
            image_bytes = StringIO()
            im = PIL.Image.open(image_path)
            im.convert('RGB').save(image_bytes, 'JPEG')
            image_bytes.seek(0)
        except IOError as e:
            logger.warning('Error reading image: {}'.format(e))
        finally:
            os.unlink(image_path)

    if image_bytes:
        logger.info('Checkin complete, adding photo. Checkin id: {}'.format(checkin_id))
        client.photos.add(params={'checkinId': checkin_id}, photo_data=image_bytes)

    return checkin_id


def _get_last_checkin(client):
    result = client.users.checkins(params={'limit': 1})
    items = result.get('checkins', {}).get('items', [])
    if items:
        return items[0]
    return None



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for pykeg.core.backup"""

import difflib
import os
import sys
import shutil
import tempfile

from pykeg.core import defaults
from pykeg.core import models

from django.core import serializers
from django.conf import settings
from django.core.files.storage import FileSystemStorage
from django.test import TransactionTestCase

from . import backup
from .testutils import make_datetime
from pykeg.core.util import get_version
from south.management.commands import migrate

from kegbot.util import kbjson

EXPECTED_SAVED_TABLES = [
    u'core_apikey',
    u'core_authenticationtoken',
    u'core_beverage',
    u'core_beverageproducer',
    u'core_controller',
    u'core_device',
    u'core_drink',
    u'core_drinkingsession',
    u'core_flowmeter',
    u'core_flowtoggle',
    u'core_invitation',
    u'core_keg',
    u'core_kegbotsite',
    u'core_kegtap',
    u'core_notificationsettings',
    u'core_picture',
    u'core_plugindata',
    u'core_stats',
    u'core_systemevent',
    u'core_thermolog',
    u'core_thermosensor',
    u'core_user',
    u'south_migrationhistory',
]


def run(cmd, args=[]):
    cmdname = cmd.__module__.split('.')[-1]
    cmd.run_from_argv([sys.argv[0], cmdname] + args)


class BackupTestCase(TransactionTestCase):
    def setUp(self):
        self.temp_storage_location = tempfile.mkdtemp(dir=os.environ.get('DJANGO_TEST_TEMP_DIR'))
        self.storage = FileSystemStorage(location=self.temp_storage_location)
        self.now = make_datetime(2014, 4, 1)

        # Backup requires a fully-populated migrations table, so inject
        # them if the unittest runner isn't actually doing them.
        if not settings.SOUTH_TESTS_MIGRATE:
            run(migrate.Command(), args=['-v', '0', '--all', '--fake'])

    def tearDown(self):
        shutil.rmtree(self.temp_storage_location)

    def load_table(self, backup_dir, table_name):
        filename = os.path.join(backup_dir, backup.TABLES_DIRNAME, table_name + '.json')
        data = open(filename, 'r').read()
        return list(serializers.deserialize('json', data))

    def assertMetadata(self, backup_dir, when=None, site_name='My Kegbot',
            num_tables=len(EXPECTED_SAVED_TABLES), num_media_files=0):
        when = when or self.now

        backup.verify_backup_directory(backup_dir)
        metadata_file = os.path.join(backup_dir, backup.METADATA_FILENAME)
        metadata_json = kbjson.loads(open(metadata_file).read())

        self.assertEqual(when, metadata_json[backup.META_CREATED_TIME])
        self.assertEquals(site_name, metadata_json[backup.META_SERVER_NAME])
        self.assertEquals(num_tables, metadata_json[backup.META_NUM_TABLES])
        self.assertEquals(num_media_files, metadata_json[backup.META_NUM_MEDIA_FILES])
        self.assertEquals(get_version(), metadata_json[backup.META_SERVER_VERSION])

    def test_create_backup_tree_empty_site(self):
        # Create a backup of an empty system.
        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            self.assertMetadata(backup_dir)

            for table_name in EXPECTED_SAVED_TABLES:
                if table_name != 'south_migrationhistory':
                    objects = self.load_table(backup_dir, table_name)
                    self.assertEqual(0, len(objects))
        finally:
            shutil.rmtree(backup_dir)

    def test_create_backup_tree_with_defaults(self):
        defaults.set_defaults(set_is_setup=True)

        site = models.KegbotSite.get()
        site.title = 'Kegbot Test 2'
        site.save()

        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            self.assertMetadata(backup_dir, site_name='Kegbot Test 2')

            sites = self.load_table(backup_dir, 'core_kegbotsite')
            self.assertEqual(1, len(sites))
            self.assertEqual('Kegbot Test 2', sites[0].object.title)

            users = self.load_table(backup_dir, 'core_user')
            self.assertEqual(1, len(users))
            self.assertEqual('guest', users[0].object.username)

            taps = self.load_table(backup_dir, 'core_kegtap')
            self.assertEqual(1, len(taps))
            self.assertEqual('Main Tap', taps[0].object.name)
        finally:
            shutil.rmtree(backup_dir)

    def test_restore_needs_erase(self):
        defaults.set_defaults(set_is_setup=True)
        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            # Restore must fail when something is already installed.
            self.assertRaises(backup.AlreadyInstalledError, backup.restore_from_directory,
                    backup_dir)

            # Erase and restore.
            backup.erase(self.storage)
            backup.restore_from_directory(backup_dir, self.storage)
        finally:
            shutil.rmtree(backup_dir)

    def test_backup_contents_same(self):
        defaults.set_defaults(set_is_setup=True)
        backup_dir = backup.create_backup_tree(self.now, storage=self.storage)

        try:
            # Verify recursive_diff works with no diffs.
            self.recursive_diff(backup_dir, backup_dir)

            second_backup_dir = backup.create_backup_tree(self.now, storage=self.storage)
            try:
                self.recursive_diff(backup_dir, second_backup_dir)
            finally:
                shutil.rmtree(second_backup_dir)
        finally:
            shutil.rmtree(backup_dir)

    def recursive_diff(self, dir1, dir2):
        dir1_files = set()
        dir2_files = set()

        for dirname, subdirs, files in os.walk(dir1):
            for filename in files:
                full_path = os.path.join(dirname, filename)
                rel_path = os.path.relpath(full_path, dir1)
                dir1_files.add(rel_path)

        for dirname, subdirs, files in os.walk(dir2):
            for filename in files:
                full_path = os.path.join(dirname, filename)
                rel_path = os.path.relpath(full_path, dir2)
                dir2_files.add(rel_path)

        self.assertEqual(dir1_files, dir2_files)

        for relfile in dir1_files:
            f1_full = os.path.join(dir1, relfile)
            f2_full = os.path.join(dir2, relfile)
            f1 = open(f1_full).read()
            f2 = open(f2_full).read()
            if f1 != f2:
                message = 'Files not equal: "{}" and "{}" differ.'.format(f1_full, f2_full)
                message += '\n' + ''.join(difflib.ndiff(f1.splitlines(True), f2.splitlines(True)))
                self.fail(message)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Routines from converting data to and from Protocol Buffer format."""

import pytz

from django.conf import settings

from kegbot.api import api_pb2
from kegbot.api import models_pb2
from kegbot.api import protoutil
from kegbot.util import util

from pykeg.core import models

_CONVERSION_MAP = {}


def converts(kind):
    def decorate(f):
        global _CONVERSION_MAP
        _CONVERSION_MAP[kind] = f
        return f
    return decorate


def datestr(dt):
    if settings.USE_TZ:
        return dt.isoformat()
    try:
        # Convert from local to UTC.
        # TODO(mikey): handle incoming datetimes with tzinfo.
        dt = util.local_to_utc(dt, settings.TIME_ZONE)
    except pytz.UnknownTimeZoneError:
        pass
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')


def ToProto(obj, full=False):
    """Converts the object to protocol format."""
    if obj is None:
        return None
    kind = obj.__class__
    if hasattr(obj, '__iter__'):
        return [ToProto(item, full) for item in obj]
    elif kind in _CONVERSION_MAP:
        return _CONVERSION_MAP[kind](obj, full)
    else:
        raise ValueError('Unknown object type: %s' % kind)


def ToDict(obj, full=False):
    res = ToProto(obj, full)
    if hasattr(res, '__iter__'):
        return [protoutil.ProtoMessageToDict(m) for m in res]
    else:
        return protoutil.ProtoMessageToDict(res)

### Model conversions


@converts(models.AuthenticationToken)
def AuthTokenToProto(record, full=False):
    ret = models_pb2.AuthenticationToken()
    ret.id = record.id
    ret.auth_device = record.auth_device
    ret.token_value = record.token_value
    if record.user_id:
        ret.username = str(record.user.username)
        ret.user.MergeFrom(ToProto(record.user))
    if record.nice_name:
        ret.nice_name = record.nice_name
    ret.created_time = datestr(record.created_time)
    ret.enabled = record.enabled
    if record.expire_time:
        ret.expire_time = datestr(record.expire_time)
    if record.pin:
        ret.pin = record.pin
    return ret


@converts(models.Picture)
def PictureToProto(record, full=False, use_png=False):
    ret = models_pb2.Image()
    if use_png:
        ret.url = record.resized_png.url
        ret.thumbnail_url = record.thumbnail_png.url
    else:
        ret.url = record.resized.url
        ret.thumbnail_url = record.thumbnail.url
    ret.original_url = record.image.url
    # TODO(mikey): This can be expensive depending on the storage backend
    # (attempts to fetch image).
    #try:
    #  ret.width = record.image.width
    #  ret.height = record.image.height
    #except IOError:
    #  pass
    if record.time:
        ret.time = datestr(record.time)
    if record.caption:
        ret.caption = record.caption
    if record.user_id:
        ret.user_id = record.user.username
    if record.keg_id:
        ret.keg_id = record.keg_id
    if record.session_id:
        ret.session_id = record.session_id
    return ret


def BeverageToBeerType(beverage, full=False):
    """Deprecated."""
    ret = models_pb2.BeerType()
    ret.id = str(beverage.id)
    ret.name = beverage.name
    ret.brewer_id = str(beverage.producer_id)
    ret.style_id = '0'
    abv = beverage.abv_percent or 0.0
    ret.abv = max(min(abv, 100.0), 0.0)
    if beverage.specific_gravity is not None:
        ret.specific_gravity = beverage.specific_gravity
    if beverage.original_gravity is not None:
        ret.original_gravity = beverage.original_gravity
    if beverage.picture:
        ret.image.MergeFrom(ToProto(beverage.picture))
    return ret


def ProducerToBrewer(producer, full=False):
    """Deprecated."""
    ret = models_pb2.Brewer()
    ret.id = str(producer.id)
    ret.name = producer.name
    if producer.country is not None:
        ret.country = producer.country
    if producer.origin_state is not None:
        ret.origin_state = producer.origin_state
    if producer.origin_city is not None:
        ret.origin_city = producer.origin_city
    if producer.production is not None:
        ret.production = 'commercial'
    if producer.url is not None:
        ret.url = producer.url
    if producer.description is not None:
        ret.description = producer.description
    if producer.picture:
        ret.image.MergeFrom(ToProto(producer.picture))
    return ret


@converts(models.Beverage)
def BeverageToProto(beverage, full=False):
    ret = models_pb2.Beverage()
    ret.id = beverage.id
    ret.name = beverage.name
    ret.producer.MergeFrom(ToProto(beverage.producer, full))
    ret.beverage_type = beverage.beverage_type

    if beverage.style:
        ret.style = beverage.style
    if beverage.description:
        ret.description = beverage.description

    if beverage.picture:
        # Use PNG for beverages.
        ret.picture.MergeFrom(
            PictureToProto(beverage.picture, full=full, use_png=True))

    if beverage.vintage_year:
        ret.vintage_year = beverage.vintage_year.year

    if beverage.abv_percent is not None:
        ret.abv_percent = beverage.abv_percent
    if beverage.calories_per_ml is not None:
        ret.calories_per_ml = beverage.calories_per_ml
    if beverage.carbs_per_ml is not None:
        ret.carbs_per_ml = beverage.carbs_per_ml

    if beverage.original_gravity is not None:
        ret.original_gravity = beverage.original_gravity
    if beverage.specific_gravity is not None:
        ret.specific_gravity = beverage.specific_gravity
    if beverage.untappd_beer_id is not None:
        ret.untappd_id = str(beverage.untappd_beer_id)

    if beverage.beverage_backend is not None:
        ret.beverage_backend = beverage.beverage_backend
    if beverage.beverage_backend_id is not None:
        ret.beverage_backend_id = beverage.beverage_backend_id

    return ret


@converts(models.BeverageProducer)
def ProducerToProto(producer, full=False):
    ret = models_pb2.BeverageProducer()
    ret.id = producer.id
    ret.name = producer.name
    if producer.country is not None:
        ret.country = producer.country
    if producer.origin_state is not None:
        ret.origin_state = producer.origin_state
    if producer.origin_city is not None:
        ret.origin_city = producer.origin_city
    ret.is_homebrew = bool(producer.is_homebrew)
    if producer.url is not None:
        ret.url = producer.url
    if producer.description is not None:
        ret.description = producer.description
    if producer.picture:
        ret.picture.MergeFrom(ToProto(producer.picture))
    return ret


@converts(models.Controller)
def ControllerToProto(controller, full=False):
    ret = models_pb2.Controller()
    ret.id = controller.id
    ret.name = controller.name
    if controller.model_name:
        ret.model_name = controller.model_name
    if controller.serial_number:
        ret.serial_number = controller.serial_number
    return ret


@converts(models.FlowMeter)
def FlowMeterToProto(flow_meter, full=False):
    ret = models_pb2.FlowMeter()
    ret.id = flow_meter.id
    ret.name = flow_meter.meter_name()
    ret.controller.MergeFrom(ToProto(flow_meter.controller))
    ret.port_name = flow_meter.port_name
    ret.ticks_per_ml = flow_meter.ticks_per_ml
    return ret


@converts(models.FlowToggle)
def FlowToggleToProto(flow_toggle, full=False):
    ret = models_pb2.FlowToggle()
    ret.id = flow_toggle.id
    ret.name = flow_toggle.toggle_name()
    ret.controller.MergeFrom(ToProto(flow_toggle.controller))
    ret.port_name = flow_toggle.port_name
    return ret


@converts(models.Drink)
def DrinkToProto(drink, full=False):
    ret = models_pb2.Drink()
    ret.id = drink.id
    ret.url = drink.get_absolute_url()
    ret.ticks = drink.ticks
    ret.volume_ml = drink.volume_ml
    ret.session_id = drink.session_id
    ret.time = datestr(drink.time)
    ret.duration = drink.duration
    if drink.keg:
        ret.keg_id = drink.keg_id
    if drink.user_id:
        ret.user_id = drink.user.username
    if drink.shout:
        ret.shout = drink.shout
    if drink.tick_time_series:
        ret.tick_time_series = drink.tick_time_series

    if full:
        if drink.user_id:
            ret.user.MergeFrom(ToProto(drink.user))
        if drink.keg:
            ret.keg.MergeFrom(ToProto(drink.keg))
        if drink.session:
            ret.session.MergeFrom(ToProto(drink.session))
        if drink.picture:
            ret.images.add().MergeFrom(ToProto(drink.picture))
    return ret


@converts(models.Keg)
def KegToProto(keg, full=False):
    ret = models_pb2.Keg()
    ret.id = keg.id
    ret.keg_type = keg.keg_type

    ret.remaining_volume_ml = float(keg.remaining_volume_ml())
    ret.full_volume_ml = keg.full_volume_ml
    ret.served_volume_ml = keg.served_volume_ml
    ret.spilled_volume_ml = keg.spilled_ml
    ret.percent_full = keg.percent_full()

    ret.start_time = datestr(keg.start_time)
    ret.end_time = datestr(keg.end_time)
    ret.online = keg.online
    if keg.description is not None:
        ret.description = keg.description

    ret.beverage.MergeFrom(ToProto(keg.type))

    # Deprecated fields.
    ret.size_id = 0
    ret.type_id = str(keg.type_id)
    ret.size_name = keg.keg_type
    ret.size_volume_ml = keg.full_volume_ml
    ret.volume_ml_remain = ret.remaining_volume_ml
    ret.spilled_ml = ret.spilled_volume_ml

    ret.type.MergeFrom(BeverageToBeerType(keg.type))
    s = models_pb2.KegSize()
    s.id = 0
    s.name = ret.size_name
    s.volume_ml = ret.size_volume_ml
    ret.size.MergeFrom(s)

    return ret


@converts(models.KegTap)
def KegTapToProto(tap, full=False):
    ret = models_pb2.KegTap()
    ret.id = tap.id
    ret.name = tap.name
    ret.sort_order = tap.sort_order
    meter = tap.current_meter()

    if meter:
        ret.meter_name = meter.meter_name()
        ret.ml_per_tick = 1 / meter.ticks_per_ml
        ret.meter.MergeFrom(ToProto(meter))
    else:
        # TODO(mikey): Remove compatibility.
        ret.meter_name = 'unknown.%s' % tap.id
        ret.ml_per_tick = 0

    toggle = tap.current_toggle()
    if toggle:
        ret.relay_name = toggle.toggle_name()
        ret.toggle.MergeFrom(ToProto(toggle))
    else:
        ret.relay_name = ''

    if tap.description is not None:
        ret.description = tap.description
    if tap.current_keg:
        ret.current_keg_id = tap.current_keg_id
        if full:
            ret.current_keg.MergeFrom(ToProto(tap.current_keg, full=True))

    if tap.temperature_sensor:
        ret.thermo_sensor.MergeFrom(ToProto(tap.temperature_sensor))
        ret.thermo_sensor_id = tap.temperature_sensor_id
        log = tap.temperature_sensor.LastLog()
        if log:
            ret.last_temperature.MergeFrom(ToProto(log))
    return ret


@converts(models.DrinkingSession)
def SessionToProto(record, full=False):
    ret = models_pb2.Session()
    ret.id = record.id
    ret.url = record.get_absolute_url()
    ret.start_time = datestr(record.start_time)
    ret.end_time = datestr(record.end_time)
    ret.volume_ml = record.volume_ml
    ret.name = record.name or ''

    if full:
        #ret.stats.MergeFrom(record.get_stats())
        ret.is_active = record.IsActiveNow()
    return ret


@converts(models.Thermolog)
def ThermoLogToProto(record, full=False):
    ret = models_pb2.ThermoLog()
    ret.id = record.id
    ret.sensor_id = record.sensor_id
    ret.temperature_c = record.temp
    ret.time = datestr(record.time)
    return ret


@converts(models.ThermoSensor)
def ThermoSensorToProto(record, full=False):
    ret = models_pb2.ThermoSensor()
    ret.id = record.id
    log = record.LastLog()
    if log:
        ret.last_log.MergeFrom(ToProto(log))

    # Deprecated
    ret.sensor_name = record.raw_name
    ret.nice_name = record.nice_name
    return ret


@converts(models.User)
def UserToProto(user, full=False):
    ret = models_pb2.User()
    ret.username = user.username
    ret.url = user.get_absolute_url()
    ret.is_active = user.is_active
    ret.display_name = user.get_full_name()
    if full:
        ret.email = user.email
        ret.is_staff = user.is_staff
        ret.is_active = user.is_active
        ret.is_superuser = user.is_superuser
        ret.last_login = datestr(user.last_login)
        ret.date_joined = datestr(user.date_joined)
    if user.mugshot_id:
        ret.image.MergeFrom(ToProto(user.mugshot))
    return ret


@converts(models.Stats)
def StatsToProto(record, full=False):
    return protoutil.DictToProtoMessage(record.stats, models_pb2.Stats())


@converts(models.SystemEvent)
def SystemEventToProto(record, full=False):
    ret = models_pb2.SystemEvent()
    ret.id = record.id
    ret.kind = record.kind
    ret.time = datestr(record.time)

    if record.drink_id:
        ret.drink_id = record.drink_id
        if full:
            ret.drink.MergeFrom(ToProto(record.drink, full=True))
    if record.keg_id:
        ret.keg_id = record.keg_id
        if full:
            ret.keg.MergeFrom(ToProto(record.keg, full=True))
    if record.session_id:
        ret.session_id = record.session_id
        if full:
            ret.session.MergeFrom(ToProto(record.session, full=True))
    if record.user_id:
        ret.user_id = str(record.user.username)
        if full:
            ret.user.MergeFrom(ToProto(record.user, full=True))

    image = None
    if record.kind in ('drink_poured', 'session_started', 'session_joined') and record.user:
        image = record.user.mugshot
    elif record.kind in ('keg_tapped', 'keg_ended'):
        if record.keg.type and record.keg.type.picture:
            image = record.keg.type.picture
    if image:
        ret.image.MergeFrom(ToProto(image))

    return ret

# Composite messages


def GetSyncResponse(active_kegs=[], active_session=[], active_users=[],
        controllers=[], drinks=[], events=[], meters=[], site_title='',
        server_version='', sound_events=[], taps=[], toggles=[]):
    ret = api_pb2.SyncResponse()
    if active_session:
        ret.active_session.MergeFrom(ToProto(active_session))

    ret.site_info.title = site_title
    ret.site_info.server_version = server_version

    for obj in active_kegs:
        ret.active_kegs.add().MergeFrom(ToProto(obj))
    for obj in active_users:
        ret.active_users.add().MergeFrom(ToProto(obj))
    for obj in controllers:
        ret.controllers.add().MergeFrom(ToProto(obj))
    for obj in drinks:
        ret.drinks.add().MergeFrom(ToProto(obj))
    for obj in events:
        ret.events.add().MergeFrom(ToProto(obj))
    for obj in meters:
        ret.meters.add().MergeFrom(ToProto(obj))
    for obj in taps:
        ret.taps.add().MergeFrom(ToProto(obj))
    for obj in toggles:
        ret.toggles.add().MergeFrom(ToProto(obj))
    return ret



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Celery tasks for Untappd."""

import requests

import foursquare

from pykeg.celery import app
from pykeg.plugin import util

logger = util.get_logger(__name__)


@app.task(name='untappd_checkin', expires=60)
def untappd_checkin(token, beer_id, timezone_name, shout=None,
        foursquare_client_id=None, foursquare_client_secret=None,
        foursquare_venue_id=None):
    logger.info('Checking in: token=%s beer_id=%s timezone_name=%s' % (
        token, beer_id, timezone_name))

    # Untappd *requires* lat/lng information when checking in with a
    # Foursquare id.  A bit silly since here we go fetching it just
    # to pass this "security" check..
    geolat = geolng = None
    if foursquare_venue_id:
        fs = foursquare.Foursquare(client_id=foursquare_client_id,
            client_secret=foursquare_client_secret)
        logger.info('Fetching venue location info from Foursquare')
        try:
            venue_info = fs.venues(foursquare_venue_id).get('venue', {})
            geolat = venue_info.get('location', {}).get('lat')
            geolng = venue_info.get('location', {}).get('lng')
            logger.info('Resolved lat/lng: {}, {}'.format(geolat, geolng))
        except foursquare.FoursquareException as e:
            logger.error('Error fetching Foursquare information: {}'.format(e))

    # TODO(mikey): API does not appear to support Olso tz names for
    # the timezone parameter.  Report as GMT, as it doesn't seem to
    # be meaningful.
    data = {
        'bid': beer_id,
        'gmt_offset': 0,
        'timezone': 'GMT',
    }

    if foursquare_venue_id and geolat and geolng:
        logger.info('Attaching Foursquare venue {}'.format(foursquare_venue_id))
        data['foursquare_id'] = foursquare_venue_id
        data['geolat'] = geolat
        data['geolng'] = geolng

    if shout:
        data['shout'] = shout

    params = {'access_token': token}
    url = 'https://api.untappd.com/v4/checkin/add'
    r = requests.post(url, params=params, data=data)
    logger.info('Response: %s' % str(r.text))



# Kegbot Server main settings file.
#
# YOU SHOULD NOT EDIT THIS FILE.  Instead, run `setup-kegbot.py` and override
# any settings in the local_settings.py file it installs.

# --------------------------------------------------------------------------- #

from __future__ import absolute_import

# Grab flags for optional modules.
from pykeg.core.optional_modules import *

import logging
from pykeg.logging.logger import RedisLogger
logging.setLoggerClass(RedisLogger)

INSTALLED_APPS = (
    'pykeg.core',
    'pykeg.web',
    'pykeg.web.api',
    'pykeg.web.account',
    'pykeg.web.kbregistration',
    'pykeg.web.kegadmin',
    'pykeg.web.kegweb',
    'pykeg.web.setup_wizard',
    'pykeg.contrib.demomode',

    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.humanize',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.staticfiles',

    'crispy_forms',
    'bootstrap-pagination',
    'imagekit',
    'gunicorn',
    'south',
)

LOGIN_REDIRECT_URL = "/account/"

KEGBOT_ADMIN_LOGIN_URL = 'auth_login'

AUTH_USER_MODEL = 'core.User'

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

### Default session serialization.
# Note: Twitter plugin requires Pickle (not JSON serializable).

SESSION_SERIALIZER = 'django.contrib.sessions.serializers.PickleSerializer'

### Kegweb specific stuff

ROOT_URLCONF = 'pykeg.web.urls'

TEMPLATE_DIRS = [
    'web/templates',
]

SITE_ID = 1

# Language code for this installation. All choices can be found here:
# http://www.w3.org/TR/REC-html40/struct/dirlang.html#langcodes
# http://blogs.law.harvard.edu/tech/stories/storyReader$15
LANGUAGE_CODE = 'en-us'

# Enable Django 1.4+ timezone support.  Do not disable this.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://media.lawrence.com/media/", "http://example.com/media/"
MEDIA_URL = '/media/'

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATIC_ROOT = ''

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = '/static/'

# URL prefix for admin static files -- CSS, JavaScript and images.
# Make sure to use a trailing slash.
# Examples: "http://foo.com/static/admin/", "/static/admin/".
ADMIN_MEDIA_PREFIX = '/static/admin/'

# Disable Django's built in host checker.
ALLOWED_HOSTS = ['*']

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
    'django.template.loaders.eggs.Loader',
)

TEMPLATE_CONTEXT_PROCESSORS = (
    'django.core.context_processors.debug',
    'django.core.context_processors.i18n',
    'django.core.context_processors.media',
    'django.core.context_processors.static',
    'django.core.context_processors.request',
    'django.contrib.auth.context_processors.auth',
    'django.contrib.messages.context_processors.messages',
    'pykeg.web.context_processors.kbsite',
)

MIDDLEWARE_CLASSES = (
    # CurrentRequest and KegbotSite middlewares added first

    'django.middleware.cache.UpdateCacheMiddleware',
    'django.middleware.gzip.GZipMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',

    'pykeg.contrib.demomode.middleware.DemoModeMiddleware',
    'pykeg.web.api.middleware.ApiRequestMiddleware',
    'pykeg.web.middleware.PrivacyMiddleware',

    'django.middleware.doc.XViewMiddleware',

    # Cache middleware should be last, except for ApiResponseMiddleWare,
    # which needs to be after it (in request order) so that it can
    # update the Cache-Control header before it (in reponse order).
    'django.middleware.cache.FetchFromCacheMiddleware',

    # ApiResponseMiddleware added last.
)

AUTHENTICATION_BACKENDS = (
    'pykeg.web.auth.local.LocalAuthBackend',
)

CACHES = {
    'default': {
        'BACKEND': 'redis_cache.cache.RedisCache',
        'LOCATION': '127.0.0.1:6379:1',
        'OPTIONS': {
            'CLIENT_CLASS': 'redis_cache.client.DefaultClient',
        }
    }
}

CACHE_MIDDLEWARE_ANONYMOUS_ONLY = True

INTERNAL_IPS = ('127.0.0.1',)

# Set to true if the database admin module should be enabled.
KEGBOT_ENABLE_ADMIN = False

# Add plugins in local_settings.py
KEGBOT_PLUGINS = [
    'pykeg.contrib.foursquare.plugin.FoursquarePlugin',
    'pykeg.contrib.twitter.plugin.TwitterPlugin',
    'pykeg.contrib.untappd.plugin.UntappdPlugin',
    'pykeg.contrib.webhook.plugin.WebhookPlugin',
]

# You probably don't want to turn this on.
DEMO_MODE = False
EMBEDDED = False

KEGBOT_BACKEND = 'pykeg.backend.backends.KegbotBackend'

### Celery

BROKER_URL = 'redis://localhost:6379/0'
CELERY_RESULT_BACKEND = 'redis://localhost:6379/0'

CELERY_QUEUES = {
    'default': {
        'exchange': 'default',
        'binding_key': 'default'
    },
    'stats': {
        'exchange': 'default',
        'binding_key': 'stats'
    },
}

CELERY_DEFAULT_QUEUE = "default"
CELERYD_CONCURRENCY = 3

from datetime import timedelta
CELERYBEAT_SCHEDULER = 'pykeg.util.celery.RedisScheduler'
CELERYBEAT_SCHEDULE = {
    'checkin': {
        'task': 'checkin',
        'schedule': timedelta(hours=12),
    }
}


### logging

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'filters': {
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'filters': ['require_debug_true'],
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
        'null': {
            'class': 'django.utils.log.NullHandler',
        },
        'redis': {
            'level': 'INFO',
            'class': 'pykeg.logging.handlers.RedisListHandler',
            'key': 'kb:log',
            'max_messages': 100,
        },
    },
    'formatters': {
        'verbose': {
            'format': '%(asctime)s %(levelname)-8s (%(name)s) %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'loggers': {
        'raven': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': False,
        },
        'pykeg': {
            'level': 'INFO',
            'handlers': ['console', 'redis'],
            'propagate': False,
        },
        '': {
            'level': 'INFO',
            'handlers': ['console'],
            'formatter': 'verbose',
        },
    },
}

### raven

if HAVE_RAVEN:
    INSTALLED_APPS += (
        'raven.contrib.django.raven_compat',
    )

    LOGGING['root']['handlers'] = ['sentry']
    LOGGING['handlers']['sentry'] = {
        'level': 'ERROR',
        'class': 'raven.contrib.django.raven_compat.handlers.SentryHandler',
    }

### django-storages
if HAVE_STORAGES:
    INSTALLED_APPS += ('storages',)

### django-nose
if HAVE_DJANGO_NOSE:
    INSTALLED_APPS += ('django_nose',)

### django.contrib.messages
MESSAGE_STORAGE = 'django.contrib.messages.storage.fallback.FallbackStorage'

### django-registration
ACCOUNT_ACTIVATION_DAYS = 3

### Statsd
STATSD_CLIENT = 'django_statsd.clients.normal'

# Set to true to route statsd pings to the debug toolbar.
KEGBOT_STATSD_TO_TOOLBAR = False

### Notifications
NOTIFICATION_BACKENDS = [
    'pykeg.notification.backends.email.EmailNotificationBackend'
]

### E-mail
EMAIL_BACKEND = 'django.core.mail.backends.dummy.EmailBackend'
EMAIL_FROM_ADDRESS = ''
EMAIL_SUBJECT_PREFIX = ''

# Bogus default values (to prevent djangofb from choking if unavailable);
# replace with site-specific values in local_settings.py, if desired.
FACEBOOK_API_KEY = ''
FACEBOOK_SECRET_KEY = ''

### Twitter

TWITTER_CONSUMER_KEY = ''
TWITTER_CONSUMER_SECRET_KEY = ''
TWITTER_REQUEST_TOKEN_URL = 'https://api.twitter.com/oauth/request_token'
TWITTER_ACCESS_TOKEN_URL = 'https://api.twitter.com/oauth/access_token'
TWITTER_AUTHORIZATION_URL = 'https://api.twitter.com/oauth/authorize'

### Foursquare

FOURSQUARE_CLIENT_ID = ''
FOURSQUARE_CLIENT_SECRET = ''
FOURSQUARE_REQUEST_PERMISSIONS = ''

### Untappd

UNTAPPD_CLIENT_ID = ''
UNTAPPD_CLIENT_SECRET = ''

### Imagekit
IMAGEKIT_DEFAULT_IMAGE_CACHE_BACKEND = 'imagekit.imagecache.NonValidatingImageCacheBackend'

TEST_RUNNER = 'pykeg.core.testutils.KegbotTestSuiteRunner'
NOSE_ARGS = ['--exe']
SKIP_SOUTH_TESTS = True
SOUTH_TESTS_MIGRATE = False

ICANHAZ_APP_DIRNAMES = ['static/jstemplates', 'jstemplates']

### Storage
DEFAULT_FILE_STORAGE = 'pykeg.web.kegweb.kbstorage.KegbotFileSystemStorage'

from pykeg.core import importhacks
try:
    from local_settings import *
except ImportError:
    msg = "Could not find local_settings.py; has Kegbot been set up?\n"
    msg += 'Tried: ' + ' '.join(importhacks.SEARCH_DIRS) + '\n\n'
    msg += 'Run setup-kegbot.py or set KEGBOT_SETTINGS_DIR to the settings directory.'
    import sys
    print>>sys.stderr, msg
    sys.exit(1)

# Override any user-specified timezone: As of Kegbot 0.9.12, this is
# specified in site settings.
TIME_ZONE = 'UTC'

# Update email addresses.
DEFAULT_FROM_EMAIL = EMAIL_FROM_ADDRESS

### socialregistration (after importing common settings)

if KEGBOT_ENABLE_ADMIN:
    INSTALLED_APPS += ('django.contrib.admin',)

### djcelery_email

if HAVE_CELERY_EMAIL:
    CELERY_EMAIL_BACKEND = EMAIL_BACKEND
    INSTALLED_APPS += ('djcelery_email',)
    EMAIL_BACKEND = 'djcelery_email.backends.CeleryEmailBackend'

### debug_toolbar

if DEBUG:
    if HAVE_DEBUG_TOOLBAR:
        INSTALLED_APPS += (
            'debug_toolbar',
        )
        MIDDLEWARE_CLASSES += (
            'debug_toolbar.middleware.DebugToolbarMiddleware',
        )
        DEBUG_TOOLBAR_PANELS = (
            'debug_toolbar.panels.version.VersionDebugPanel',
            'debug_toolbar.panels.timer.TimerDebugPanel',
            'debug_toolbar.panels.settings.SettingsPanel',
            'debug_toolbar.panels.headers.HeaderDebugPanel',
            'debug_toolbar.panels.request_vars.RequestVarsDebugPanel',
            'debug_toolbar.panels.template.TemplateDebugPanel',
            'debug_toolbar.panels.sql.SQLDebugPanel',
            'debug_toolbar.panels.signals.SignalDebugPanel',
            'debug_toolbar.panels.logging.LoggingPanel',
            #'debug_toolbar.panels.profiling.ProfilingDebugPanel',
        )
        if HAVE_MEMCACHE_TOOLBAR:
            INSTALLED_APPS += ('debug_toolbar_memcache',)
            if HAVE_MEMCACHE:
                DEBUG_TOOLBAR_PANELS += ('debug_toolbar_memcache.panels.memcache.MemcachePanel',)
            elif HAVE_PYLIBMC:
                DEBUG_TOOLBAR_PANELS += ('debug_toolbar_memcache.panels.pylibmc.PylibmcPanel',)

# Add all plugin template dirs to search path.
from pykeg.core.util import get_plugin_template_dirs
TEMPLATE_DIRS += get_plugin_template_dirs(KEGBOT_PLUGINS)

### Statsd

# Needs SECRET_KEY so must be imported after local settings.

STATSD_PATCHES = [
    'django_statsd.patches.db',
    'django_statsd.patches.cache',
]

if HAVE_STATSD:
    MIDDLEWARE_CLASSES = (
        'django_statsd.middleware.GraphiteRequestTimingMiddleware',
        'django_statsd.middleware.GraphiteMiddleware',
    ) + MIDDLEWARE_CLASSES

    INSTALLED_APPS += ('django_statsd',)

if DEBUG and HAVE_DEBUG_TOOLBAR and KEGBOT_STATSD_TO_TOOLBAR:
    MIDDLEWARE_CLASSES = (
        'debug_toolbar.middleware.DebugToolbarMiddleware',
    ) + MIDDLEWARE_CLASSES
    DEBUG_TOOLBAR_PANELS = (
        'django_statsd.panel.StatsdPanel',
    ) + DEBUG_TOOLBAR_PANELS
    STATSD_CLIENT = 'django_statsd.clients.toolbar'

### First/last middlewares.

MIDDLEWARE_CLASSES = (
    'pykeg.web.middleware.CurrentRequestMiddleware',
    'pykeg.web.middleware.KegbotSiteMiddleware',
) + MIDDLEWARE_CLASSES

MIDDLEWARE_CLASSES += ('pykeg.web.api.middleware.ApiResponseMiddleware',)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for registration functions."""

from django.core import mail
from django.test import TestCase
from django.test.utils import override_settings
from pykeg.backend import get_kegbot_backend
from pykeg.core import models as core_models
from pykeg.core import defaults


class ForgotPasswordTest(TestCase):
    def setUp(self):
        self.backend = get_kegbot_backend()
        defaults.set_defaults(set_is_setup=True)

        self.user = core_models.User.objects.create(username='notification_user',
            email='test@example.com')

        # Password reset requires a usable password.
        self.user.set_password('1234')
        self.user.save()

    @override_settings(EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend')
    @override_settings(EMAIL_FROM_ADDRESS='test-from@example')
    def test_notifications(self):
        response = self.client.get('/accounts/password/reset/')
        self.assertContains(response, "Reset Password", status_code=200)
        self.assertEquals(0, len(mail.outbox))

        response = self.client.post('/accounts/password/reset/', data={'email': 'test@example.com'},
            follow=True)
        self.assertContains(response, 'E-Mail Sent', status_code=200)
        self.assertEquals(1, len(mail.outbox))

        msg = mail.outbox[0]

        self.assertEquals('Password reset on My Kegbot', msg.subject)
        self.assertEquals(['test@example.com'], msg.to)
        self.assertEquals('test-from@example', msg.from_email)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for Twitter plugin."""

from django.test import TransactionTestCase
from pykeg.contrib.twitter import plugin
from pykeg.plugin.datastore import InMemoryDatastore

from pykeg.core import models


class TwitterPluginTestCase(TransactionTestCase):

    def setUp(self):
        self.datastore = InMemoryDatastore(plugin.TwitterPlugin.get_short_name())
        self.plugin = plugin.TwitterPlugin(datastore=self.datastore)
        self.user = models.User.objects.create(username='twitter_test')

    def test_plugin(self):
        self.assertEqual({}, self.plugin.get_site_profile())
        self.assertEqual({}, self.plugin.get_user_profile(self.user))

        fake_profile = {
            plugin.KEY_OAUTH_TOKEN: '1',
            plugin.KEY_OAUTH_TOKEN_SECRET: '2',
            plugin.KEY_TWITTER_NAME: '3',
            plugin.KEY_TWITTER_ID: '4',
        }
        self.plugin.save_site_profile('1', '2', '3', '4')
        self.assertEquals(fake_profile, self.plugin.get_site_profile())
        self.plugin.remove_site_profile()
        self.assertEquals({}, self.plugin.get_site_profile())
        self.plugin.save_site_profile('1', '2', '3', '4')

    def test_get_site_twitter_settings_form(self):
        settings_form = self.plugin.get_site_twitter_settings_form()
        expected = {
            'tweet_keg_events': True,
            'tweet_session_events': True,
            'tweet_drink_events': False,
            'include_guests': True,
            'include_pictures': False,
            'append_url': True,
            'keg_started_template': 'Woot! Just tapped a new keg of $BEER!',
            'keg_ended_template': 'The keg of $BEER has been finished.',
            'session_started_template': '$DRINKER kicked off a new session on $SITENAME.',
            'session_joined_template': '$DRINKER joined the session.',
            'drink_poured_template': '$DRINKER poured $VOLUME of $BEER on $SITENAME.',
            'user_drink_poured_template': 'Just poured $VOLUME of $BEER on $SITENAME. #kegbot',
        }
        self.assertEquals(expected, dict(settings_form.initial))

    def test_truncate_tweet(self):
        msg = 'x' * 138 + 'x'
        truncated = plugin.truncate_tweet(msg)
        self.assertEquals(msg, truncated)

        longer_msg = 'x' * 138 + ' xxxx'
        truncated = plugin.truncate_tweet(longer_msg)
        self.assertEqual(139, len(truncated))
        self.assertEqual('x' * 138 + plugin.TRUNCATE_STR, truncated)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Data storage interface and implementations for plugins."""

from kegbot.util.util import AttrDict
from pykeg.core import models


class PluginDatastore(object):
    """Interface for plugins to persist plugin-specific data."""

    def __init__(self, plugin_name):
        self.plugin_name = plugin_name

    def set(self, key, value):
        """Store plugin-specific data."""
        raise NotImplementedError

    def get(self, key, default=None):
        """Fetch plugin-specific data."""
        raise NotImplementedError

    def delete(self, key):
        raise NotImplementedError

    def save_form(self, form, prefix):
        """Helper method to save a form using the specified per-field prefix."""
        for field_name, value in form.cleaned_data.iteritems():
            self.set('%s:%s' % (prefix, field_name), value)

    def load_form(self, form_cls, prefix, form_kwargs={}):
        """Helper method to load a form using the specified per-field prefix."""
        data = AttrDict()
        for field_name, field in form_cls.base_fields.iteritems():
            initial = self.get('%s:%s' % (prefix, field_name))
            if initial is not None:
                data[field_name] = field.to_python(initial)
            else:
                if field.initial is not None:
                    data[field_name] = field.initial
        return form_cls(initial=data, **form_kwargs)


class ModelDatastore(PluginDatastore):
    """PluginDatastore backed by models.PluginData rows."""

    def set(self, key, value):
        if value is None:
            self.delete(key)
            return
        try:
            row = models.PluginData.objects.get(plugin_name=self.plugin_name, key=key)
            row.value = value
            row.save()
        except models.PluginData.DoesNotExist:
            models.PluginData.objects.create(plugin_name=self.plugin_name, key=key,
                value=value)

    def get(self, key, default=None):
        try:
            row = models.PluginData.objects.get(plugin_name=self.plugin_name,
                key=key)
            return row.value
        except models.PluginData.DoesNotExist:
            return default

    def delete(self, key):
        try:
            models.PluginData.objects.get(plugin_name=self.plugin_name,
                key=key).delete()
        except models.PluginData.DoesNotExist:
            pass


class InMemoryDatastore(PluginDatastore):
    def __init__(self, *args, **kwargs):
        super(InMemoryDatastore, self).__init__(*args, **kwargs)
        self.data = {}

    def _keyname(self, key):
        """Returns the datastore-namespaced key name."""
        return '{}:{}'.format(self.plugin_name, key)

    def set(self, key, value):
        if value is None:
            self.delete(key)
        else:
            self.data[self._keyname(key)] = value

    def get(self, key, default=None):
        return self.data.get(self._keyname(key), default)

    def delete(self, key):
        keyname = self._keyname(key)
        if keyname in self.data:
            del self.data[keyname]



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import urlparse

from django.conf import settings
from django.core.files.storage import FileSystemStorage

try:
    from storages.backends.s3boto import S3BotoStorage
except ImportError:
    S3BotoStorage = None

from pykeg.backend import get_kegbot_backend

S3_STATIC_BUCKET = getattr(settings, 'S3_STATIC_BUCKET', None)


class KegbotFileSystemStorage(FileSystemStorage):
    """Default storage backend that crafts absolute urls from KegbotSite.base_url.

    Since the storage backed is not a singleton within django request
    processing (and thus there's not a single object we can pre-configure
    with the base URL), this custom backend seems necessary.
    """

    def url(self, name):
        be = get_kegbot_backend()
        base_url = be.get_base_url()
        if not self.base_url.startswith(base_url):
            self.base_url = urlparse.urljoin(base_url, self.base_url)
        return super(KegbotFileSystemStorage, self).url(name)

if S3BotoStorage:
    class S3StaticStorage(S3BotoStorage):
        """Uses settings.S3_STATIC_BUCKET instead of AWS_STORAGE_BUCKET_NAME."""

        def __init__(self, *args, **kwargs):
            kwargs['bucket'] = S3_STATIC_BUCKET
            super(S3StaticStorage, self).__init__(*args, **kwargs)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Celery tasks for Twitter."""

import os

import tweepy
from pykeg.celery import app
from pykeg.core.util import download_to_tempfile

from pykeg.plugin import util

logger = util.get_logger(__name__)


@app.task(name='twitter_tweet', expires=60)
def send_tweet(consumer_key, consumer_secret, oauth_token, oauth_token_secret, tweet, image_url):
    auth = tweepy.OAuthHandler(consumer_key, consumer_secret, secure=True)
    auth.set_access_token(oauth_token, oauth_token_secret)
    api = tweepy.API(auth)

    local_media = None
    if image_url:
        try:
            local_media = download_to_tempfile(image_url)
        except IOError as e:
            logger.warning('Error downloading media url "%s": %s' % (image_url, e))

    if not local_media:
        logger.info('Sending text-only tweet: {}'.format(repr(tweet)))
        result = api.update_status(tweet)
    else:
        logger.info('Sending media tweet.')
        try:
            result = api.update_with_media(local_media, status=tweet)
        finally:
            os.unlink(local_media)
    return result



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegbot authentication backend interface."""

from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured


def get_auth_backend():
    """Returns the sole authentication backend."""
    backends = auth.get_backends()
    if len(backends) != 1:
        raise ImproperlyConfigured('Expected exactly 1 backend (found {})'.format(
            len(backends)))
    return backends[0]


class AuthException(Exception):
    """Top-level exception class."""


class UserExistsException(AuthException):
    """Registration error indicating this username/email is taken."""


class AuthBackend(object):

    ### Django methods

    def authenticate(self, **credentials):
        raise NotImplementedError

    def get_user(self, user_id):
        raise NotImplementedError

    ### Kegbot methods

    def is_manager(self, user):
        """Returns true if this user has manager privileges.

        True implies `is_member()` is also True.
        """
        raise NotImplementedError

    def set_is_manager(self, user, is_manager):
        """Sets or clears the manager status for this user."""
        raise NotImplementedError

    def is_owner(self, user):
        """Returns true if this user has owner privileges.

        True implies `is_manager()` is also True.
        """
        raise NotImplementedError

    def set_is_owner(self, user, is_owner):
        """Sets or clears the manager status for this user."""
        raise NotImplementedError

    def register(self, email, username, password=None, photo=None):
        """Registers a new user on this system.

        Args:
            email: new user's email address.
            username: desired username
            password: initial password
            photo: photo data

        Returns:
            The new User instance on success.
        """
        raise NotImplementedError



#!/usr/bin/env python
#
# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import zipfile

import redis

from operator import itemgetter

from django.conf import settings
from django.contrib import messages
from pykeg.web.decorators import staff_member_required
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.core.files.storage import get_storage_class
from django.db.models import Q
from django.http import Http404
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils import timezone
from django.views.decorators.http import require_http_methods

from kegbot.util import kbjson

from pykeg.web.api import devicelink
from pykeg.celery import app as celery_app
from pykeg.core import backup
from pykeg.core import checkin
from pykeg.core import models
from pykeg.logging.handlers import RedisListHandler
from pykeg.util.email import build_message

from pykeg.web.kegadmin import forms

import logging
logger = logging.getLogger(__name__)


@staff_member_required
def dashboard(request):
    context = RequestContext(request)

    # Hack: Schedule an update checkin if it looks like it's been a while.
    # This works around sites that are not running celerybeat.
    checkin.schedule_checkin()

    email_backend = getattr(settings, 'EMAIL_BACKEND', None)
    email_configured = email_backend and email_backend != 'django.core.mail.backends.dummy.EmailBackend'
    email_configured = email_configured and bool(getattr(settings, 'EMAIL_FROM_ADDRESS', None))

    context['email_configured'] = email_configured

    if settings.BROKER_URL.startswith('redis:'):
        try:
            r = redis.StrictRedis.from_url(settings.BROKER_URL)
            r.ping()
        except redis.RedisError as e:
            context['redis_error'] = e.message if e.message else "Unknown error."

    checkin_info = checkin.get_last_checkin()
    context['last_checkin_time'] = checkin_info[0]
    context['checkin'] = checkin_info[1]

    active_users = models.User.objects.filter(is_active=True).exclude(username='guest')
    context['num_users'] = len(active_users)

    recent_time = timezone.now() - datetime.timedelta(days=30)
    new_users = models.User.objects.filter(date_joined__gte=recent_time).exclude(username='guest')
    context['num_new_users'] = len(new_users)

    return render_to_response('kegadmin/dashboard.html', context_instance=context)


@staff_member_required
def general_settings(request):
    context = RequestContext(request)
    kbsite = request.kbsite

    form = forms.GeneralSiteSettingsForm(instance=kbsite)

    if request.method == 'POST':
        form = forms.GeneralSiteSettingsForm(request.POST, instance=kbsite)
        if form.is_valid():
            form.save()
            messages.success(request, 'Settings were updated.')
    context['settings_form'] = form

    return render_to_response('kegadmin/index.html', context_instance=context)


@staff_member_required
def location_settings(request):
    context = RequestContext(request)
    kbsite = request.kbsite

    form = forms.LocationSiteSettingsForm(instance=kbsite)

    if request.method == 'POST':
        form = forms.LocationSiteSettingsForm(request.POST, instance=kbsite)
        if form.is_valid():
            form.save()
            messages.success(request, 'Settings were updated.')
    context['settings_form'] = form

    return render_to_response('kegadmin/index.html', context_instance=context)


@staff_member_required
def advanced_settings(request):
    context = RequestContext(request)
    kbsite = request.kbsite

    form = forms.AdvancedSiteSettingsForm(instance=kbsite)

    if request.method == 'POST':
        form = forms.AdvancedSiteSettingsForm(request.POST, instance=kbsite)
        if form.is_valid():
            form.save()
            guest_image = request.FILES.get('guest_image')
            if guest_image:
                pic = models.Picture.objects.create()
                pic.image.save(guest_image.name, guest_image)
                pic.save()
                kbsite.guest_image = pic
                kbsite.save()
            messages.success(request, 'Settings were updated.')
    context['settings_form'] = form

    return render_to_response('kegadmin/index.html', context_instance=context)


@staff_member_required
def email(request):
    context = RequestContext(request)
    kbsite = request.kbsite

    email_backend = getattr(settings, 'EMAIL_BACKEND', None)
    email_configured = email_backend and email_backend != 'django.core.mail.backends.dummy.EmailBackend'
    email_configured = email_configured and bool(getattr(settings, 'EMAIL_FROM_ADDRESS', None))

    if request.method == 'POST':
        if 'send_test_email' in request.POST:
            test_email_form = forms.TestEmailForm(request.POST)
            if test_email_form.is_valid():
                address = test_email_form.cleaned_data.get('address')
                context['site_name'] = kbsite.title
                context['site_url'] = kbsite.base_url()
                context['settings_url'] = context['site_url'] + '/account'
                message = build_message(address, 'notification/email_test.html', context)
                message.send(fail_silently=True)
                messages.success(request, 'E-mail successfully sent to %s' % address)

    context['email_configured'] = email_configured

    return render_to_response('kegadmin/email.html', context_instance=context)


@staff_member_required
def export(request):
    context = RequestContext(request)
    backups = []
    storage = get_storage_class()()

    if request.method == 'POST':
        if 'package_backup' in request.POST:
            request.backend.build_backup()
            messages.success(request, 'The backup is being generated; please reload in a few.')

        elif 'delete_backup' in request.POST:
            backup_file = os.path.normpath(os.path.basename(request.POST['backup_name']))
            backup_file = os.path.join(backup.BACKUPS_DIRNAME, backup_file)
            if storage.exists(backup_file):
                storage.delete(backup_file)
                messages.success(request, 'Backup deleted.')
            else:
                messages.warning(request, 'Unknown backup file.')

    if storage.exists(backup.BACKUPS_DIRNAME):
        subdirs, files = storage.listdir(backup.BACKUPS_DIRNAME)
        for filename in files:
            if filename[-3:] == 'zip':
                storage_filename = os.path.join(backup.BACKUPS_DIRNAME, filename)
                with storage.open(storage_filename, mode='rb') as backup_file:
                    archive = zipfile.ZipFile(backup_file)
                    metadata = backup.read_metadata(archive)
                    metadata['size_bytes'] = storage.size(storage_filename)
                    metadata['url'] = storage.url(storage_filename)
                    metadata['backup_name'] = filename
                    backups.append(metadata)

    backups.sort(key=itemgetter(backup.META_CREATED_TIME), reverse=True)
    context['backups'] = backups

    return render_to_response('kegadmin/backup_export.html', context_instance=context)


@staff_member_required
def workers(request):
    context = RequestContext(request)

    try:
        inspector = celery_app.control.inspect()
        pings = inspector.ping() or {}
        stats = inspector.stats() or {}
        queues = inspector.active_queues() or {}
    except redis.RedisError as e:
        context['error'] = e

    status = {}
    if not pings and 'error' not in context:
        context['error'] = 'No response from workers. Not running?'
    else:
        for k, v in pings.iteritems():
            status[k] = {
                'status': 'ok' if v.get('ok') else 'unknown',
            }
        for k, v in stats.iteritems():
            if k in status:
                status[k]['stats'] = v
        for k, v in queues.iteritems():
            if k in status:
                status[k]['active_queues'] = v

    context['status'] = status
    context['raw_stats'] = kbjson.dumps(context['status'], indent=2)

    return render_to_response('kegadmin/workers.html', context_instance=context)


@staff_member_required
def controller_list(request):
    context = RequestContext(request)
    context['controllers'] = models.Controller.objects.all()
    return render_to_response('kegadmin/controller_list.html', context_instance=context)


@staff_member_required
def controller_detail(request, controller_id):
    controller = get_object_or_404(models.Controller, id=controller_id)
    delete_controller_form = forms.DeleteControllerForm()
    add_flow_meter_form = forms.AddFlowMeterForm()
    add_flow_meter_form.fields['controller'].initial = controller_id
    add_flow_toggle_form = forms.AddFlowToggleForm()
    add_flow_toggle_form.fields['controller'].initial = controller_id
    context = RequestContext(request)

    if request.method == 'POST':
        if 'delete_controller' in request.POST:
            delete_controller_form = forms.DeleteControllerForm(request.POST)
            if delete_controller_form.is_valid():
                controller.delete()
                messages.success(request, 'The controller was deleted.')
                return redirect('kegadmin-controllers')
        elif 'add_flow_meter' in request.POST:
            add_flow_meter_form = forms.AddFlowMeterForm(request.POST)
            if add_flow_meter_form.is_valid():
                add_flow_meter_form.save()
                messages.success(request, 'Flow Meter added successfully.')
                return redirect('kegadmin-controllers')
        elif 'edit_flow_meter' in request.POST:
            flowmeter = models.FlowMeter.objects.filter(id=request.POST.get('flowmeter_id'))
            flowmeter.update(port_name=request.POST.get('port_name'))
            flowmeter.update(ticks_per_ml=request.POST.get('ticks_per_ml'))
            messages.success(request, 'Flow Meter successfully updated.')
            return redirect('kegadmin-controllers')
        elif 'delete_flow_meter' in request.POST:
            flowmeter = models.FlowMeter.objects.filter(id=request.POST.get('flowmeter_id')).delete()
            messages.success(request, 'Flow Meter removed successfully.')
            return redirect('kegadmin-controllers')
        elif 'add_flow_toggle' in request.POST:
            add_flow_toggle_form = forms.AddFlowToggleForm(request.POST)
            if add_flow_toggle_form.is_valid():
                add_flow_toggle_form.save()
                messages.success(request, 'Flow Toggle added successfully.')
                return redirect('kegadmin-controllers')
        elif 'edit_flow_toggle' in request.POST:
            flowtoggle = models.FlowToggle.objects.filter(id=request.POST.get('flowtoggle_id'))
            flowtoggle.update(port_name=request.POST.get('port_name'))
            messages.success(request, 'Flow Toggle successfully updated.')
            return redirect('kegadmin-controllers')
        elif 'delete_flow_toggle' in request.POST:
            flowmeter = models.FlowToggle.objects.filter(id=request.POST.get('flowtoggle_id')).delete()
            messages.success(request, 'Flow Toggle removed successfully.')
            return redirect('kegadmin-controllers')

    context['controller'] = controller
    context['delete_controller_form'] = delete_controller_form
    context['add_flow_meter_form'] = add_flow_meter_form
    context['add_flow_toggle_form'] = add_flow_toggle_form
    return render_to_response('kegadmin/controller_detail.html', context_instance=context)


@staff_member_required
def tap_list(request):
    context = RequestContext(request)
    context['taps'] = models.KegTap.objects.all()
    return render_to_response('kegadmin/tap_list.html', context_instance=context)


@staff_member_required
def add_tap(request):
    context = RequestContext(request)
    form = forms.TapForm()
    if request.method == 'POST':
        form = forms.TapForm(request.POST)
        if form.is_valid():
            form.save()
            messages.success(request, 'Tap created.')
            return redirect('kegadmin-taps')
    context['form'] = form
    return render_to_response('kegadmin/add_tap.html', context_instance=context)


@staff_member_required
def tap_detail(request, tap_id):
    tap = get_object_or_404(models.KegTap, id=tap_id)
    available_kegs = models.Keg.objects.filter(online=False, finished=False).order_by('id')

    record_drink_form = forms.RecordDrinkForm()
    activate_keg_form = forms.ChangeKegForm()
    tap_settings_form = forms.TapForm(instance=tap)

    if request.method == 'POST':
        if 'submit_change_keg_form' in request.POST:
            activate_keg_form = forms.ChangeKegForm(request.POST)
            if activate_keg_form.is_valid():
                activate_keg_form.save(tap)
                messages.success(request, 'The new keg was activated. Bottoms up!')
                return redirect('kegadmin-taps')

        if 'submit_keg_choice' in request.POST:
            keg_id = request.POST.get('keg_id')
            keg = models.Keg.objects.get(id=keg_id)
            d = request.backend.attach_keg(tap, keg)
            messages.success(request, 'The new keg was activated. Bottoms up!')
            return redirect('kegadmin-taps')

        elif 'submit_tap_form' in request.POST:
            tap_settings_form = forms.TapForm(request.POST, instance=tap)
            if tap_settings_form.is_valid():
                tap_settings_form.save()
                messages.success(request, 'Tap settings saved.')
                tap_settings_form = forms.TapForm(instance=tap)

        elif 'submit_delete_tap_form' in request.POST:
            delete_form = forms.DeleteTapForm(request.POST)
            if delete_form.is_valid():
                if tap.current_keg:
                    request.backend.end_keg(tap)
                tap.delete()
                messages.success(request, 'Tap deleted.')
                return redirect('kegadmin-taps')

        elif 'submit_end_keg_form' in request.POST:
            end_keg_form = forms.EndKegForm(request.POST)
            if end_keg_form.is_valid():
                old_keg = request.backend.end_keg(tap)
                messages.success(request, 'Keg %s was ended.' % old_keg.id)

        elif 'submit_record_drink' in request.POST:
            record_drink_form = forms.RecordDrinkForm(request.POST)
            if record_drink_form.is_valid():
                user = record_drink_form.cleaned_data.get('user')
                volume_ml = record_drink_form.cleaned_data.get('volume_ml')
                d = request.backend.record_drink(tap, ticks=0, username=user,
                    volume_ml=volume_ml)
                messages.success(request, 'Drink %s recorded.' % d.id)
            else:
                messages.error(request, 'Please enter a valid volume and user.')

        elif 'submit_record_spill' in request.POST:
            record_drink_form = forms.RecordDrinkForm(request.POST)
            if record_drink_form.is_valid():
                user = record_drink_form.cleaned_data.get('user')
                volume_ml = record_drink_form.cleaned_data.get('volume_ml')
                d = request.backend.record_drink(tap, ticks=0, username=user,
                    volume_ml=volume_ml, spilled=True)
                messages.success(request, 'Spill recorded.')
            else:
                messages.error(request, 'Please enter a valid volume.')

        else:
            messages.warning(request, 'No form data was found. Bug?')

    end_keg_form = forms.EndKegForm(initial={'keg': tap.current_keg})

    context = RequestContext(request)
    context['tap'] = tap
    context['current_keg'] = tap.current_keg
    context['available_kegs'] = available_kegs
    context['activate_keg_form'] = activate_keg_form
    context['record_drink_form'] = record_drink_form
    context['end_keg_form'] = end_keg_form
    context['tap_settings_form'] = tap_settings_form
    context['delete_tap_form'] = forms.DeleteTapForm()
    return render_to_response('kegadmin/tap_detail.html', context_instance=context)


@staff_member_required
def keg_list(request):
    context = RequestContext(request)
    kegs = models.Keg.objects.all().order_by('-id')
    paginator = Paginator(kegs, 25)

    page = request.GET.get('page')
    try:
        kegs = paginator.page(page)
    except PageNotAnInteger:
        kegs = paginator.page(1)
    except EmptyPage:
        kegs = paginator.page(paginator.num_pages)

    context['kegs'] = kegs
    return render_to_response('kegadmin/keg_list.html', context_instance=context)


@staff_member_required
def keg_detail(request, keg_id):
    keg = get_object_or_404(models.Keg, id=keg_id)

    form = forms.EditKegForm(instance=keg)
    if request.method == 'POST':
        form = forms.EditKegForm(request.POST, instance=keg)
        if form.is_valid():
            form.save()
            messages.success(request, 'Keg updated.')
            return redirect('kegadmin-kegs')

    context = RequestContext(request)
    context['keg'] = keg
    context['remaining'] = keg.remaining_volume_ml()
    context['form'] = form
    return render_to_response('kegadmin/keg_detail.html', context_instance=context)


@staff_member_required
def keg_add(request):
    add_keg_form = forms.KegForm()
    if request.method == 'POST':
        if 'submit_add_keg' in request.POST:
            add_keg_form = forms.KegForm(request.POST)
            if add_keg_form.is_valid():
                add_keg_form.save()
                messages.success(request, 'New keg added.')
                return redirect('kegadmin-kegs')

    context = RequestContext(request)
    context['keg'] = 'new'
    context['form'] = add_keg_form
    return render_to_response('kegadmin/keg_add.html', context_instance=context)


@staff_member_required
def user_list(request):
    context = RequestContext(request)

    if request.method == 'POST':
        form = forms.FindUserForm(request.POST)
        if form.is_valid():
            username = form.cleaned_data.get('username')
            try:
                user = models.User.objects.get(username=username)
                return redirect('kegadmin-edit-user', user.id)
            except models.User.DoesNotExist:
                messages.error(request, 'User "%s" does not exist.' % username)

    users = models.User.objects.exclude(username='guest').order_by('-id')
    paginator = Paginator(users, 25)

    page = request.GET.get('page')
    try:
        users = paginator.page(page)
    except PageNotAnInteger:
        users = paginator.page(1)
    except EmptyPage:
        users = paginator.page(paginator.num_pages)

    context['users'] = users
    return render_to_response('kegadmin/user_list.html', context_instance=context)


@staff_member_required
def add_user(request):
    context = RequestContext(request)
    form = forms.UserForm()
    if request.method == 'POST':
        form = forms.UserForm(request.POST)
        if form.is_valid():
            instance = form.save(commit=False)
            instance.set_password(form.cleaned_data.get('password'))
            instance.save()
            messages.success(request, 'User "%s" created.' % instance.username)
            return redirect('kegadmin-users')
    context['form'] = form
    return render_to_response('kegadmin/add_user.html', context_instance=context)


@staff_member_required
def user_detail(request, user_id):
    edit_user = get_object_or_404(models.User, id=user_id)
    context = RequestContext(request)
    profile_form = forms.UserProfileForm(instance=edit_user)

    if request.method == 'POST':
        if 'submit_enable' in request.POST:
            if edit_user.is_active:
                messages.error(request, 'User is already enabled.')
            else:
                edit_user.is_active = True
                edit_user.save()
                messages.success(request, 'User %s was enabled.' % edit_user.username)

        elif 'submit_disable' in request.POST:
            if edit_user.is_guest():
                messages.error(request, 'Cannot disable the guest user.')
            elif not edit_user.is_active:
                messages.error(request, 'User is already disabled.')
            else:
                edit_user.is_active = False
                edit_user.save()
                messages.success(request, 'User %s was disabled.' % edit_user.username)

        elif 'submit_add_staff' in request.POST:
            if edit_user.is_staff:
                messages.error(request, 'User is already staff.')
            else:
                edit_user.is_staff = True
                edit_user.save()
                messages.success(request, 'User %s staff status enabled.' % edit_user.username)

        elif 'submit_remove_staff' in request.POST:
            if edit_user.is_guest():
                messages.error(request, 'Cannot change staff status on the guest user.')
            elif not edit_user.is_staff:
                messages.error(request, 'User is not currently staff.')
            else:
                edit_user.is_staff = False
                edit_user.save()
                messages.success(request, 'User %s staff status disabled.' % edit_user.username)

        elif 'submit_update_profile' in request.POST:
            profile_form = forms.UserProfileForm(request.POST, request.FILES, instance=edit_user)
            if profile_form.is_valid():
                profile_form.save()
                messages.success(request, 'User %s e-profile updated' % edit_user.username)

        else:
            messages.error(request, 'Unknown form submitted.')

    context['profile_form'] = profile_form
    context['edit_user'] = edit_user
    context['tokens'] = edit_user.tokens.all().order_by('created_time')

    return render_to_response('kegadmin/user_detail.html', context_instance=context)


@staff_member_required
def drink_list(request):
    context = RequestContext(request)
    drinks = models.Drink.objects.all().order_by('-time')
    paginator = Paginator(drinks, 25)

    page = request.GET.get('page')
    try:
        drinks = paginator.page(page)
    except PageNotAnInteger:
        drinks = paginator.page(1)
    except EmptyPage:
        drinks = paginator.page(paginator.num_pages)

    context['drinks'] = drinks
    return render_to_response('kegadmin/drink_list.html', context_instance=context)


@staff_member_required
@require_http_methods(["POST"])
def drink_edit(request, drink_id):
    drink = get_object_or_404(models.Drink, id=drink_id)

    if 'submit_cancel' in request.POST:
        form = forms.CancelDrinkForm(request.POST)
        old_keg = drink.keg
        if form.is_valid():
            request.backend.cancel_drink(drink)
            messages.success(request, 'Drink %s was cancelled.' % drink_id)
            return redirect(old_keg.get_absolute_url())
        else:
            messages.error(request, 'Invalid request')
            return redirect(drink.get_absolute_url())

    if 'submit_spill' in request.POST:
        form = forms.CancelDrinkForm(request.POST)
        old_keg = drink.keg
        if form.is_valid():
            request.backend.cancel_drink(drink, spilled=True)
            messages.success(request, 'Drink %s was spilled.' % drink_id)
            return redirect(old_keg.get_absolute_url())
        else:
            messages.error(request, 'Invalid request')
            return redirect(drink.get_absolute_url())

    elif 'submit_reassign' in request.POST:
        form = forms.ReassignDrinkForm(request.POST)
        if form.is_valid():
            username = form.cleaned_data.get('username', None)
            try:
                request.backend.assign_drink(drink, username)
                messages.success(request, 'Drink %s was reassigned.' % drink_id)
            except models.User.DoesNotExist:
                messages.error(request, 'No such user')
        else:
            messages.error(request, 'Invalid request')
        return redirect(drink.get_absolute_url())

    elif 'submit_edit_volume' in request.POST:
        form = forms.ChangeDrinkVolumeForm(request.POST)
        if form.is_valid():
            volume_ml = form.cleaned_data.get('volume_ml')
            if volume_ml == drink.volume_ml:
                messages.warning(request, 'Drink volume unchanged.')
            else:
                request.backend.set_drink_volume(drink, volume_ml)
                messages.success(request, 'Drink %s was updated.' % drink_id)
        else:
            messages.error(request, 'Please provide a valid volume.')
        return redirect(drink.get_absolute_url())

    messages.error(request, 'Unknown action.')
    return redirect(drink.get_absolute_url())


@staff_member_required
def token_list(request):
    context = RequestContext(request)
    tokens = models.AuthenticationToken.objects.all().order_by('-created_time')
    paginator = Paginator(tokens, 25)

    page = request.GET.get('page')
    try:
        tokens = paginator.page(page)
    except PageNotAnInteger:
        tokens = paginator.page(1)
    except EmptyPage:
        tokens = paginator.page(paginator.num_pages)

    context['tokens'] = tokens
    return render_to_response('kegadmin/token_list.html', context_instance=context)


@staff_member_required
def token_detail(request, token_id):
    token = get_object_or_404(models.AuthenticationToken, id=token_id)

    username = ''
    if token.user:
        username = token.user.username

    form = forms.TokenForm(instance=token, initial={'username': username})
    if request.method == 'POST':
        form = forms.TokenForm(request.POST, instance=token)
        if form.is_valid():
            instance = form.save(commit=False)
            instance.user = form.cleaned_data['user']
            instance.save()
            messages.success(request, 'Token updated.')

    context = RequestContext(request)
    context['token'] = token
    context['form'] = form
    return render_to_response('kegadmin/token_detail.html', context_instance=context)


@staff_member_required
def add_token(request):
    context = RequestContext(request)
    form = forms.AddTokenForm()
    if request.method == 'POST':
        form = forms.AddTokenForm(request.POST)
        if form.is_valid():
            instance = form.save(commit=False)
            instance.user = form.cleaned_data['user']
            instance.save()
            messages.success(request, 'Token created.')
            return redirect('kegadmin-tokens')
    context['form'] = form
    return render_to_response('kegadmin/add_token.html', context_instance=context)


@staff_member_required
def beverages_list(request):
    context = RequestContext(request)
    beers = models.Beverage.objects.all().order_by('name')
    paginator = Paginator(beers, 25)

    page = request.GET.get('page')
    try:
        beers = paginator.page(page)
    except PageNotAnInteger:
        beers = paginator.page(1)
    except EmptyPage:
        beers = paginator.page(paginator.num_pages)

    context['beverages'] = beers
    return render_to_response('kegadmin/beer_type_list.html', context_instance=context)


@staff_member_required
def beverage_detail(request, beer_id):
    btype = get_object_or_404(models.Beverage, id=beer_id)

    form = forms.BeverageForm(instance=btype)
    if request.method == 'POST':
        form = forms.BeverageForm(request.POST, instance=btype)
        if form.is_valid():
            btype = form.save()
            new_image = request.FILES.get('new_image')
            if new_image:
                pic = models.Picture.objects.create()
                pic.image.save(new_image.name, new_image)
                pic.save()
                btype.picture = pic
                btype.save()

            messages.success(request, 'Beer type updated.')
            return redirect('kegadmin-beverages')

    context = RequestContext(request)
    context['beer_type'] = btype
    context['form'] = form
    return render_to_response('kegadmin/beer_type_detail.html', context_instance=context)


@staff_member_required
def beverage_add(request):

    form = forms.BeverageForm()
    if request.method == 'POST':
        form = forms.BeverageForm(request.POST)
        if form.is_valid():
            btype = form.save()
            new_image = request.FILES.get('new_image')
            if new_image:
                pic = models.Picture.objects.create()
                pic.image.save(new_image.name, new_image)
                pic.save()
                btype.picture = pic
                btype.save()

            messages.success(request, 'Beer type added.')
            return redirect('kegadmin-beverages')

    context = RequestContext(request)
    context['beer_type'] = 'new'
    context['form'] = form
    return render_to_response('kegadmin/beer_type_add.html', context_instance=context)


@staff_member_required
def beverage_producer_list(request):
    context = RequestContext(request)
    brewers = models.BeverageProducer.objects.all().order_by('name')
    paginator = Paginator(brewers, 25)

    page = request.GET.get('page')
    try:
        brewers = paginator.page(page)
    except PageNotAnInteger:
        brewers = paginator.page(1)
    except EmptyPage:
        brewers = paginator.page(paginator.num_pages)

    context['brewers'] = brewers
    return render_to_response('kegadmin/brewer_list.html', context_instance=context)


@staff_member_required
def beverage_producer_detail(request, brewer_id):
    brewer = get_object_or_404(models.BeverageProducer, id=brewer_id)

    form = forms.BeverageProducerForm(instance=brewer)
    if request.method == 'POST':
        form = forms.BeverageProducerForm(request.POST, instance=brewer)
        if form.is_valid():
            form.save()
            messages.success(request, 'Brewer updated.')
            return redirect('kegadmin-beverage-producers')

    context = RequestContext(request)
    context['brewer'] = brewer
    context['form'] = form
    return render_to_response('kegadmin/brewer_detail.html', context_instance=context)


@staff_member_required
def beverage_producer_add(request):
    form = forms.BeverageProducerForm()
    if request.method == 'POST':
        form = forms.BeverageProducerForm(request.POST)
        if form.is_valid():
            btype = form.save()
            new_image = request.FILES.get('new_image')
            if new_image:
                pic = models.Picture.objects.create()
                pic.image.save(new_image.name, new_image)
                pic.save()
                btype.picture = pic
                btype.save()

            messages.success(request, 'Brewer added.')
            return redirect('kegadmin-beverage-producers')

    context = RequestContext(request)
    context['brewer'] = 'new'
    context['form'] = form
    return render_to_response('kegadmin/brewer_add.html', context_instance=context)


@staff_member_required
def autocomplete_beverage(request):
    search = request.GET.get('q')
    if search:
        beverages = models.Beverage.objects.filter(Q(name__icontains=search) | Q(producer__name__icontains=search))
    else:
        beverages = models.Beverage.objects.all()
    beverages = beverages[:10]  # autocomplete widget limited to 10
    values = []
    for beverage in beverages:
        values.append({
            'name': beverage.name,
            'id': beverage.id,
            'producer_name': beverage.producer.name,
            'producer_id': beverage.producer.id,
            'style': beverage.style,
        })
    return HttpResponse(kbjson.dumps(values, indent=None),
      mimetype='application/json', status=200)


@staff_member_required
def autocomplete_user(request):
    search = request.GET.get('q')
    if search:
        users = models.User.objects.filter(Q(username__icontains=search) | Q(email__icontains=search) | Q(display_name__icontains=search))
    else:
        users = models.User.objects.all()
    users = users[:10]  # autocomplete widget limited to 10
    values = []
    for user in users:
        values.append({
            'username': user.username,
            'id': user.id,
            'email': user.email,
            'display_name': user.get_full_name(),
            'is_active': user.is_active,
        })
    return HttpResponse(kbjson.dumps(values, indent=None),
      mimetype='application/json', status=200)


@staff_member_required
def autocomplete_token(request):
    search = request.GET.get('q')
    if search:
        tokens = models.AuthenticationToken.objects.filter(
            Q(token_value__icontains=search) | Q(nice_name__icontains=search))
    else:
        tokens = models.AuthenticationToken.objects.all()
    tokens = tokens[:10]  # autocomplete widget limited to 10
    values = []
    for token in tokens:
        values.append({
            'username': token.user.username,
            'id': token.id,
            'auth_device': token.auth_device,
            'token_value': token.token_value,
            'enabled': token.enabled,
        })
    return HttpResponse(kbjson.dumps(values, indent=None),
      mimetype='application/json', status=200)


@staff_member_required
def plugin_settings(request, plugin_name):
    plugin = request.plugins.get(plugin_name, None)
    if not plugin:
        raise Http404('Plugin "%s" not loaded' % plugin_name)

    view = plugin.get_admin_settings_view()
    if not view:
        raise Http404('No settings for this plugin')

    return view(request, plugin)


@staff_member_required
def logs(request):
    context = RequestContext(request)
    handlers = logger.parent.handlers

    logs = []
    for h in handlers:
        if isinstance(h, RedisListHandler):
            logs = list(h.get_logs())
            logs.reverse()
            break

    context['logs'] = logs
    return render_to_response('kegadmin/logs.html', context_instance=context)


@staff_member_required
def link_device(request):
    context = RequestContext(request)
    form = forms.LinkDeviceForm()
    if request.method == 'POST':
        form = forms.LinkDeviceForm(request.POST)
        if form.is_valid():
            code = form.cleaned_data.get('code')
            try:
                status = devicelink.confirm_link(code)
                name = status.get('name', 'New device')
                messages.success(request, '{} linked!'.format(name))
            except devicelink.LinkExpiredException:
                messages.error(request, 'Code incorrect or expired.')
            return redirect('kegadmin-link-device')
    context['form'] = form
    return render_to_response('kegadmin/link_device.html', context_instance=context)



from django.conf.urls import patterns
from django.conf.urls import url

from pykeg.web.account.views import password_change
from pykeg.web.account.views import password_change_done


urlpatterns = patterns('pykeg.web.account.views',
    url(r'^$', 'account_main', name='kb-account-main'),
    url(r'^activate/(?P<activation_key>[0-9a-zA-Z]+)/$', 'activate_account', name='activate-account'),
    url(r'^password/done/$', password_change_done, name='password_change_done'),
    url(r'^password/$', password_change, name='password_change'),
    url(r'^profile/$', 'edit_profile', name='account-profile'),
    url(r'^invite/$', 'invite', name='account-invite'),
    url(r'^notifications/$', 'notifications', name='account-notifications'),
    url(r'^regenerate-api-key/$', 'regenerate_api_key', name='regen-api-key'),
    url(r'^plugin/(?P<plugin_name>\w+)/$', 'plugin_settings', name='account-plugin-settings'),
)

from pykeg.plugin import util
urlpatterns += util.get_account_urls()



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Constants about physical keg shell."""

# Most common shell sizes, from smalles to largest.
CORNY = 'corny'
SIXTH_BARREL = 'sixth'
EURO_HALF_BARREL = 'euro-half'
QUARTER_BARREL = 'quarter'
EURO = 'euro'
HALF_BARREL = 'half-barrel'
OTHER = 'other'

VOLUMES_ML = {
    CORNY: 18927.1,
    SIXTH_BARREL: 19570.6,
    EURO_HALF_BARREL: 50000.0,
    QUARTER_BARREL: 29336.9,
    EURO: 100000.0,
    HALF_BARREL: 58673.9,
    OTHER: 0.0
}

DESCRIPTIONS = {
    CORNY: 'Corny Keg (5 gal)',
    SIXTH_BARREL: 'Sixth Barrel (5.17 gal)',
    EURO_HALF_BARREL: 'European Half Barrel (50 L)',
    QUARTER_BARREL: 'Quarter Barrel (7.75 gal)',
    EURO: 'European Full Barrel (100 L)',
    HALF_BARREL: 'Half Barrel (15.5 gal)',
    OTHER: 'Other',
}

CHOICES = [(x, DESCRIPTIONS[x]) for x in reversed(sorted(VOLUMES_ML, key=VOLUMES_ML.get))]


def find_closest_keg_size(volume_ml, tolerance_ml=100.0):
    """Returns the nearest fuzzy match name within tolerance_ml.

    If no match is found, OTHER is returned.
    """
    for size_name, size_volume_ml in VOLUMES_ML.iteritems():
        diff = abs(volume_ml - size_volume_ml)
        if diff <= tolerance_ml:
            return size_name
    return OTHER


def get_description(keg_type):
    return DESCRIPTIONS.get(keg_type, 'Unknown')



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import messages
from django.core.urlresolvers import reverse
from django.shortcuts import redirect
from socialregistration.clients.oauth import OAuthError
from pykeg.web.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import render_to_response
from django.template import RequestContext

from httplib2 import HttpLib2Error

from . import forms
from . import tasks

# Session keys for temporarily storing oauth client.
SESSION_KEY_SITE_TWITTER = 'site-twitter'
SESSION_KEY_USER_TWITTER = 'user-twitter'


@staff_member_required
def admin_settings(request, plugin):
    context = RequestContext(request)

    consumer_key, consumer_secret = plugin.get_credentials()
    initial = {
        'consumer_key': consumer_key or '',
        'consumer_secret': consumer_secret or '',
    }

    credentials_form = forms.CredentialsForm(initial=initial)
    settings_form = plugin.get_site_twitter_settings_form()
    tweet_form = forms.SendTweetForm()

    if request.method == 'POST':
        if 'submit-keys' in request.POST:
            credentials_form = forms.CredentialsForm(request.POST)
            if credentials_form.is_valid():
                consumer_key = credentials_form.cleaned_data['consumer_key']
                consumer_secret = credentials_form.cleaned_data['consumer_secret']
                plugin.set_credentials(consumer_key, consumer_secret)
                messages.success(request, 'Keys updated')

        elif 'submit-settings' in request.POST:
            settings_form = forms.SiteSettingsForm(request.POST)
            if settings_form.is_valid():
                plugin.save_form(settings_form, 'site_settings')
                messages.success(request, 'Settings updated')

        elif 'submit-tweet' in request.POST:
            tweet_form = forms.SendTweetForm(request.POST)
            if tweet_form.is_valid():
                tweet = tweet_form.cleaned_data['tweet_custom']
                consumer_key, consumer_secret = plugin.get_credentials()
                profile = plugin.get_site_profile()
                if consumer_key and consumer_secret and profile:
                    oauth_token = profile.get('oauth_token')
                    oauth_token_secret = profile.get('oauth_token_secret')
                    tasks.send_tweet(consumer_key, consumer_secret, oauth_token, oauth_token_secret,
                        tweet, image_url=None)
                    messages.success(request, 'Tweet sent')
                else:
                    messages.error(request, 'There was a problem sending tweet')

    context['have_credentials'] = consumer_key and consumer_secret
    context['plugin'] = plugin
    context['site_profile'] = plugin.get_site_profile()
    context['credentials_form'] = credentials_form
    context['settings_form'] = settings_form
    context['tweet_form'] = tweet_form

    return render_to_response('contrib/twitter/admin_settings.html', context_instance=context)


@login_required
def user_settings(request, plugin):
    context = RequestContext(request)
    user = request.user

    consumer_key, consumer_secret = plugin.get_credentials()

    settings_form = plugin.get_user_settings_form(user)

    if request.method == 'POST':
        if 'submit-settings' in request.POST:
            settings_form = forms.UserSettingsForm(request.POST)
            if settings_form.is_valid():
                plugin.save_user_settings_form(user, settings_form)
                messages.success(request, 'Settings updated')

    context['have_credentials'] = consumer_key and consumer_secret
    context['plugin'] = plugin
    context['profile'] = plugin.get_user_profile(user)
    context['settings_form'] = settings_form

    return render_to_response('contrib/twitter/twitter_user_settings.html', context_instance=context)


@staff_member_required
def site_twitter_redirect(request):
    if 'submit-remove' in request.POST:
        plugin = request.plugins.get('twitter')
        plugin.remove_site_profile()
        messages.success(request, 'Removed Twitter account.')
        return redirect('kegadmin-plugin-settings', plugin_name='twitter')

    plugin = request.plugins['twitter']

    client = plugin.get_client()
    url = request.build_absolute_uri(reverse('plugin-twitter-site_twitter_callback'))
    client.set_callback_url(url)

    return do_redirect(request, client, 'kegadmin-plugin-settings', SESSION_KEY_SITE_TWITTER)


@staff_member_required
def site_twitter_callback(request):
    plugin = request.plugins.get('twitter')
    client = request.session.get(SESSION_KEY_SITE_TWITTER)

    if not client:
        messages.error(request, 'Lost OAuth session')
    else:
        del request.session[SESSION_KEY_SITE_TWITTER]
        try:
            token = client.complete(dict(request.GET.items()))
        except KeyError:
            messages.error(request, 'Session expired.')
        except OAuthError, error:
            messages.error(request, str(error))
        else:
            user_info = client.get_user_info()
            plugin = request.plugins.get('twitter')
            plugin.save_site_profile(token.key, token.secret, user_info['screen_name'],
                int(user_info['user_id']))
            messages.success(request, 'Successfully linked to @%s' % user_info['screen_name'])

    return redirect('kegadmin-plugin-settings', plugin_name='twitter')


@login_required
def user_twitter_redirect(request):
    if 'submit-remove' in request.POST:
        plugin = request.plugins.get('twitter')
        plugin.remove_user_profile(request.user)
        messages.success(request, 'Removed Twitter account.')
        return redirect('account-plugin-settings', plugin_name='twitter')

    plugin = request.plugins['twitter']
    client = plugin.get_client()
    url = request.build_absolute_uri(reverse('plugin-twitter-user_twitter_callback'))
    client.set_callback_url(url)

    return do_redirect(request, client, 'account-plugin-settings', SESSION_KEY_USER_TWITTER)


@login_required
def user_twitter_callback(request):
    plugin = request.plugins['twitter']
    client = request.session.get(SESSION_KEY_USER_TWITTER)

    if not client:
        messages.error(request, 'Lost OAuth session')
    else:
        del request.session[SESSION_KEY_USER_TWITTER]
        try:
            token = client.complete(dict(request.GET.items()))
        except KeyError:
            messages.error(request, 'Session expired.')
        except OAuthError, error:
            messages.error(request, str(error))
        else:
            user_info = client.get_user_info()
            plugin = request.plugins.get('twitter')
            plugin.save_user_profile(request.user, token.key, token.secret,
                user_info['screen_name'], int(user_info['user_id']))
            messages.success(request, 'Successfully linked to @%s' % user_info['screen_name'])

    return redirect('account-plugin-settings', plugin_name='twitter')


def do_redirect(request, client, next_url_name, session_key):
    """Common redirect method, handling any incidental errors.

    Args:
        request: the incoming request
        client: the TwitterClient context_instance
        next_url_name: Django URL name to redirect to upon error

    Returns:
        A redirect response, either to the OAuth redirect URL
        or to `next_url_name` on error.
    """
    request.session[session_key] = client
    try:
        return redirect(client.get_redirect_url())
    except OAuthError, e:
        messages.error(request, 'Error: %s' % str(e))
        return redirect(next_url_name, plugin_name='twitter')
    except HttpLib2Error, e:
        # This path can occur when api.twitter.com is unresolvable
        # or unreachable.
        messages.error(request, 'Twitter API server not available. Try again later. (%s)' % str(e))
        return redirect(next_url_name, plugin_name='twitter')



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.http import HttpResponse
from django.utils.cache import add_never_cache_headers

from . import util

import logging

LOGGER = logging.getLogger(__name__)

# These paths are allowed without authorization, regardless of
# site privacy settings.
WHITELISTED_API_PATHS = (
    '/api/devices/link',
    '/api/v1/devices/link',
    '/api/login',
    '/api/v1/login',
    '/api/version',
    '/api/v1/version',
    '/api/get-api-key',
    '/api/v1/get-api-key',
)

WHITELISTED_API_PATHS += getattr(settings, 'KEGBOT_EXTRA_WHITELISTED_API_PATHS', ())


class ApiRequestMiddleware:
    def process_view(self, request, view_func, view_args, view_kwargs):
        request.is_kb_api_request = util.is_api_request(request)
        if not request.is_kb_api_request:
            # Not an API request. Skip me.
            return None

        try:
            if request.need_setup:
                raise ValueError('Setup required')
            elif request.need_upgrade:
                raise ValueError('Upgrade required')

            need_auth = util.needs_auth(view_func)
            privacy = request.kbsite.privacy
            if request.path.startswith(WHITELISTED_API_PATHS):
                # API request to whitelisted path.
                need_auth = False
            else:
                # API request to non-whitelisted path, in non-public site privacy mode.
                # Demand API key.
                if privacy == 'members' and not request.user.is_authenticated():
                    need_auth = True
                elif privacy == 'staff' and not request.user.is_staff:
                    need_auth = True

            if need_auth:
                util.check_api_key(request)

            return None

        except Exception, e:
            return util.wrap_exception(request, e)


class ApiResponseMiddleware:
    def process_exception(self, request, exception):
        """Wraps exceptions for API requests."""
        if util.is_api_request(request):
            return util.wrap_exception(request, exception)
        return None

    def process_response(self, request, response):
        if not util.is_api_request(request):
            return response

        if not isinstance(response, HttpResponse):
            data = util.prepare_data(response)
            data['meta'] = {
                'result': 'ok'
            }
            response = util.build_response(request, data, 200)

        add_never_cache_headers(response)
        return response


def cache_key(request):
    return 'api:%s' % request.get_full_path()



"""Defines HAVE_* flags for several optional packages."""

import imp

try:
    imp.find_module('debug_toolbar')
    HAVE_DEBUG_TOOLBAR = True
except ImportError:
    HAVE_DEBUG_TOOLBAR = False

try:
    imp.find_module('raven.contrib.django')
    HAVE_RAVEN = True
except ImportError:
    HAVE_RAVEN = False

try:
    imp.find_module('django_statsd')
    HAVE_STATSD = True
except ImportError:
    HAVE_STATSD = False

try:
    imp.find_module('storages')
    HAVE_STORAGES = True
except ImportError:
    HAVE_STORAGES = False

try:
    imp.find_module('memcache')
    HAVE_MEMCACHE = True
except ImportError:
    HAVE_MEMCACHE = False

try:
    imp.find_module('pylibmc')
    HAVE_PYLIBMC = True
except ImportError:
    HAVE_PYLIBMC = False

try:
    imp.find_module('debug_toolbar_memcache')
    HAVE_MEMCACHE_TOOLBAR = True
except ImportError:
    HAVE_MEMCACHE_TOOLBAR = False

try:
    imp.find_module('djcelery_email')
    HAVE_CELERY_EMAIL = True
except ImportError:
    HAVE_CELERY_EMAIL = False

try:
    imp.find_module('django_nose')
    HAVE_DJANGO_NOSE = True
except ImportError:
    HAVE_DJANGO_NOSE = False
